/*
 *  Copyright (c) by Allin Cottrell
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

/* gui_utils.c for gretl */

#include "gretl.h"
#include <sys/stat.h>

char rcfile[MAXLEN];
char *storelist = NULL;

static GtkWidget *help_view = NULL;
int db_is_native = 0;

extern GtkItemFactoryEntry view_items[3];
extern GtkTooltips *gretl_tips;
extern int session_saved;
extern GtkWidget *console_view;
extern GtkWidget *console_dialog;
extern GtkWidget *mysheet;
extern GtkWidget *toolbar_box;
extern char *space_to_score (char *str);

extern char want_toolbar[6];
extern char calculator[MAXSTR];
extern char editor[MAXSTR];
extern char Rcommand[MAXSTR];

/* filelist stuff */
#define MAXRECENT 4

char *endbit (char *dest, char *src, int addscore);
static void printfilelist (int filetype, FILE *fp);

static char datalist[MAXRECENT][MAXSTR], *datap[MAXRECENT];
static char sessionlist[MAXRECENT][MAXSTR], *sessionp[MAXRECENT];
static char scriptlist[MAXRECENT][MAXSTR], *scriptp[MAXRECENT];

/* helpfile vars */
static int help_length, gui_help_length, script_help_length;

/* searching stuff */
static int look_for_string (char *haystack, char *needle, int nStart);
static void close_find_dialog (GtkWidget *widget, gpointer data);
static void find_in_help (GtkWidget *widget, gpointer data);
static void find_in_clist (GtkWidget *widget, gpointer data);
static void cancel_find (GtkWidget *widget, gpointer data);
static void find_string_dialog (void (*YesFunc)(), void (*NoFunc)(),
				gpointer data);
static void set_up_viewer_menu (GtkWidget *window, windata_t *vwin, 
				GtkItemFactoryEntry items[], int msize);
static GtkWidget *find_window = NULL;
static GtkWidget *find_entry;
static char *needle;

static void make_prefs_tab (GtkWidget *notebook, int tab);
static void apply_changes (GtkWidget *widget, gpointer data);
static void read_rc_file (void);

extern void do_coeff_intervals (gpointer data, guint i, GtkWidget *w);
extern void save_plot (char *fname, GPT_SPEC *plot);
extern gboolean console_handler (GtkWidget *w, GdkEventKey *key, 
				 gpointer user_data);
extern void do_panel_diagnostics (gpointer data, guint u, GtkWidget *w);

typedef struct {
    char *key,         /* config file variable name */
         *description; /* How the field will show up in the options dialog */
    char *var;         /* string variable */
    char type;         /* 'C' for string, 'I' for integer */
    short tab;         /* which tab (if any) does the item fall under? */
    GtkWidget *widget;
} RCVARS;

RCVARS rc_vars[] = {
    {"userdir", "User's gretl directory", paths.userdir, 'C', 1, NULL},
    {"gretldir", "Main gretl directory", paths.gretldir, 'C', 1, NULL},
    {"Rcommand", "Command to launch GNU R", Rcommand, 'C', 1, NULL},
    {"expert", "Expert mode (no warnings)", expert, 'I', 1, NULL},
    {"updater", "Tell me about gretl updates", updater, 'I', 1, NULL},
    {"binbase", "gretl database directory", paths.binbase, 'C', 2, NULL},
    {"ratsbase", "RATS data directory", paths.ratsbase, 'C', 2, NULL},
    {"dbhost_ip", "Database server IP", paths.dbhost_ip, 'C', 2, NULL},
    {"calculator", "Calculator", calculator, 'C', 3, NULL},
    {"editor", "Editor", editor, 'C', 3, NULL},
    {"toolbar", "Show gretl toolbar", want_toolbar, 'I', 3, NULL},
    {"fontspec", "Fixed font", fontspec, 'C', 0, NULL},
    {"", "", NULL, 0, 0, NULL},    
};

GtkItemFactoryEntry model_items[] = {
    { "/_File", NULL, NULL, 0, "<Branch>" },
    { "/File/_Save as text...", NULL, file_save, SAVE_MODEL, NULL },
    { "/File/Save to session as icon...", NULL, gretl_callback, 
      STORE_MODEL, NULL },
    { "/File/Save as icon and close", NULL, quick_remember_model, 0, NULL },
    { "/_Edit", NULL, NULL, 0, "<Branch>" },
    { "/Edit/_Copy selection", NULL, text_copy, 0, NULL },
    { "/Edit/Copy _all", NULL, text_copy, 1, NULL },
    { "/_Tests", NULL, NULL, 0, "<Branch>" },    
    { "/Tests/omit variables", NULL, gretl_callback, OMIT, NULL },
    { "/Tests/add variables", NULL, gretl_callback, ADD, NULL },
    { "/Tests/sep1", NULL, NULL, 0, "<Separator>" },
    { "/Tests/non-linearity (squares)", NULL, do_lmtest, AUX_SQ, NULL },
    { "/Tests/non-linearity (logs)", NULL, do_lmtest, AUX_LOG, NULL },
    { "/Tests/sep2", NULL, NULL, 0, "<Separator>" },
    { "/Tests/autocorrelation", NULL, do_lmtest, AUX_AR, NULL },
    { "/Tests/heteroskedasticity", NULL, do_lmtest, AUX_WHITE, NULL },
    { "/Tests/Chow test", NULL, gretl_callback, CHOW, NULL },
    { "/Tests/CUSUM test", NULL, do_cusum, 0, NULL },
    { "/Tests/ARCH", NULL, gretl_callback, ARCH, NULL },
    { "/Tests/normality of residual", NULL, do_resid_freq, 0, NULL },
    { "/Tests/panel diagnostics", NULL, do_panel_diagnostics, 0, NULL },
    { "/_Graphs", NULL, NULL, 0, "<Branch>" }, 
    { "/Graphs/residual plot", NULL, NULL, 0, "<Branch>" },
    { "/Graphs/fitted, actual plot", NULL, NULL, 0, "<Branch>" },
    { "/_Model data", NULL, NULL, 0, "<Branch>" },
    { "/_Model data/Display actual, fitted, residual", NULL, 
      display_fit_resid, 0, NULL },
    { "/_Model data/Forecasts with standard errors", NULL, 
      gretl_callback, FCAST, NULL },
    { "/_Model data/Confidence intervals for coefficients", NULL, 
      do_coeff_intervals, 0, NULL },
    { "/_Model data/Add to data set/fitted values", NULL, 
      fit_resid_callback, 1, NULL },
    { "/_Model data/Add to data set/residuals", NULL, 
      fit_resid_callback, 0, NULL },
    { "/_Model data/Add to data set/squared residuals", NULL, 
      fit_resid_callback, 2, NULL },
    { "/_Model data/Add to data set/error sum of squares", NULL, 
      model_stat_callback, ESS, NULL },
    { "/_Model data/Add to data set/standard error of residuals", NULL, 
      model_stat_callback, SIGMA, NULL },
    { "/_Model data/Add to data set/R-squared", NULL, 
      model_stat_callback, R2, NULL },
    { "/_Model data/Add to data set/T*R-squared", NULL, 
      model_stat_callback, TR2, NULL },
    { "/_Model data/Add to data set/degrees of freedom", NULL, 
      model_stat_callback, DF, NULL },
    { "/_Model data/coefficient covariance matrix", NULL, 
      do_outcovmx, 0, NULL },
    { "/_Model data/sep1", NULL, NULL, 0, "<Separator>" },
    { "/_Model data/Define new variable...", NULL, gretl_callback, 
      MODEL_GENR, NULL },
    { "/_LaTeX", NULL, NULL, 0, "<Branch>" },
    { "/LaTeX/view tabular", NULL, view_latex, 0, NULL },
    { "/LaTeX/view equation", NULL, view_latex, 1, NULL },
    { "/LaTeX/save tabular", NULL, file_save, SAVE_TEX_TAB, NULL },
    { "/LaTeX/save equation", NULL, file_save, SAVE_TEX_EQ, NULL },
};

GtkItemFactoryEntry help_items[] = {
    { "/_Topics", NULL, NULL, 0, "<Branch>" },    
    { "/Topics/Generate variable syntax", NULL, do_help, GENR, NULL },
    { "/Topics/sep1", NULL, NULL, 0, "<Separator>" },
    { "/Topics/Graphing", NULL, do_help, GNUPLOT, NULL },
    { "/Topics/sep2", NULL, NULL, 0, "<Separator>" },
    { "/Topics/Estimation/Ordinary Least Squares", 
      NULL, do_help, OLS, NULL },
    { "/Topics/Estimation/Weighted Least Squares", 
      NULL, do_help, WLS, NULL },
    { "/Topics/Estimation/Cochrane-Orcutt", NULL, do_help, CORC, NULL },
    { "/Topics/Estimation/HCCM", NULL, do_help, HCCM, NULL },
    { "/Topics/Estimation/Hildreth-Lu", NULL, do_help, HILU, NULL },
    { "/Topics/Estimation/Heteroskedasticity", NULL, do_help, HSK, NULL },
    { "/Topics/Estimation/Autoregressive estimation", NULL, do_help, AR, NULL },
    { "/Topics/Estimation/Vector Autoregression", NULL, do_help, VAR, NULL },
    { "/Topics/Estimation/Two-Stage Least Squares", NULL, do_help, TSLS, NULL }, 
    { "/Topics/Estimation/_Logit", NULL, do_help, LOGIT, NULL }, 
    { "/Topics/Estimation/_Probit", NULL, do_help, PROBIT, NULL }, 
    { "/Topics/Estimation/_Rank Correlation", NULL, do_help, SPEARMAN, NULL }, 
    { "/Topics/sep3", NULL, NULL, 0, "<Separator>" },
    { "/Topics/Hypothesis tests/omit variables", NULL, do_help, OMIT, NULL },
    { "/Topics/Hypothesis tests/add variables", NULL, do_help, ADD, NULL },
    { "/Topics/Hypothesis tests/LM test", NULL, do_help, LMTEST, NULL },
    { "/Topics/Hypothesis tests/Dickey-Fuller test", 
      NULL, do_help, ADF, NULL },
    { "/Topics/Hypothesis tests/Chow test", NULL, do_help, CHOW, NULL },
    { "/Topics/Hypothesis tests/Cointegration test", NULL, do_help, 
      COINT, NULL },
    { "/_Find", NULL, menu_find, 0, NULL }
};

GtkItemFactoryEntry script_help_items[] = {
    { "/_Find", NULL, menu_find, 0, NULL }
};

/* ........................................................... */

int copyfile (const char *src, const char *dest) 
{
    FILE *srcfd, *destfd;
    char buf[8192];
    size_t n;
   
    if ((srcfd = fopen(src, "rb")) == NULL) {
	return 1; 
    }
    if ((destfd = fopen(dest, "wb")) == NULL) {
	fclose(srcfd);
	return 1;
    }
    while ((n = fread(buf, 1, sizeof buf, srcfd)) > 0) {
	fwrite(buf, 1, n, destfd);
    }
    fclose(srcfd);
    fclose(destfd);
    return 0;
}

/* ........................................................... */

int isdir (const char *path)
{
    struct stat buf;

    if (stat(path, &buf) == 0 && S_ISDIR(buf.st_mode)) 
	return 1;
    else 
	return 0;
}

/* ........................................................... */

int getbufline (char *buf, char *line, int init)
{
    static int pos;
    int i = 0;

    if (init) pos = 0;
    else {
	while (buf[i+pos] != '\n') {
	    line[i] = buf[i+pos];
	    if (buf[i+pos] == 0)
		return 0;
	    i++;
	}
	pos += i + 1;
	line[i] = 0;
    }
    return i;
}

/* ........................................................... */

void append_dir (char *fname, const char *dir)
{
    if (dir != NULL) strcat(fname, dir);
    strcat(fname, SLASHSTR);
}

/* ........................................................... */

void set_rcfile (void) 
{
    char *tmp;

    tmp = getenv("HOME");
    strcpy(rcfile, tmp);
    strcat(rcfile, "/.gretlrc");
    read_rc_file(); 
}

/* ........................................................... */

void set_win_rcfile (PATHS *ppaths) 
{
    strcpy(rcfile, ppaths->userdir);
    strcat(rcfile, "gretl.rc");
    read_rc_file(); 
}

/* ........................................................... */

static void delete_file (GtkWidget *widget, char *fle) 
{
    remove(fle);
    g_free(fle);
}

/* ........................................................... */

static void delete_file_viewer (GtkWidget *widget, gpointer data) 
{
    gtk_widget_destroy((GtkWidget *) data);
}

/* ........................................................... */

void delete_model (GtkWidget *widget, gpointer data) 
{
    MODEL *pmod = (MODEL *) data;
    if (pmod->name == NULL) {
	free_model(pmod);
	pmod = NULL;
    }
}

/* ........................................................... */

void delete_widget (GtkWidget *widget, gpointer data)
{
    gtk_widget_destroy(data);
}

/* ........................................................... */

void catch_key (GtkWidget *w, GdkEventKey *key)
{
    if (key->keyval == GDK_q) 
        gtk_widget_destroy(w);
    else if (key->keyval == GDK_s) 
	quick_remember_model 
	    (gtk_object_get_data(GTK_OBJECT(w), "ddata"), 0, NULL);
}

/* ........................................................... */

void *mymalloc (size_t size) 
{
    void *mem;
   
    if((mem = g_malloc(size)) == NULL) 
	errbox("Out of memory!");
    return mem;
}

/* ........................................................... */

void *myrealloc (void *ptr, size_t size) 
{
    void *mem;
   
    if((mem = g_realloc(ptr, size)) == NULL) 
	errbox("Out of memory!");
    return mem;
}

/* ........................................................... */

void clear_clist (GtkWidget *widget)
{
    gtk_clist_clear(GTK_CLIST(widget));
    if (popup_connected) {
	gtk_signal_disconnect_by_func(GTK_OBJECT(mdata->listbox),
				      (GtkSignalFunc) main_popup, 
				      NULL);
	popup_connected = 0;
    }
}

/* ........................................................... */

void register_data (const char *fname)
{    
    char datacmd[MAXLEN];

    /* basic accounting */
    data_file_open = 1;
    orig_vars = datainfo->v;

    /* sync main window with datafile */
    populate_clist(mdata->listbox, datainfo);
    set_sample_label(datainfo);
    menubar_state(TRUE);
    session_state(TRUE);    

    /* record opening of data file in command log */
    if (fname != NULL) {
	mkfilelist(1, fname);
	sprintf(datacmd, "open %s", fname);
	check_cmd(datacmd);
	cmd_init(datacmd); 
    }   
}

/* ........................................................... */

void do_open_data (GtkWidget *w, gpointer data)
     /* cases: 
	- called from dialog: user has said Yes to opening data file,
	although a data file is already open
	- reached without dialog, in expert mode or when no datafile
	is open yet
     */
{
    gint datatype, err;
    print_t prn;
    dialog_t *d = NULL;
    windata_t *fwin = NULL;

    if (data != NULL) {    
	if (w == NULL) { /* not coming from edit_dialog */
	    fwin = (windata_t *) data;
	} else {
	    d = (dialog_t *) data;
	    fwin = (windata_t *) d->data;
	}
    }

    /* check file type first */
    if (bufopen(&prn)) return;
    datatype = detect_filetype(paths.datfile, &paths, &prn);
    prnclose(&prn);

    if (datatype == 2) {
	do_open_csv_box(paths.datfile, OPEN_CSV);
	return;
    }
    else if (datatype == 3) {
	do_open_csv_box(paths.datfile, OPEN_BOX);
	return;
    }
    else 	
	err = get_data(&Z, datainfo, &paths,
		       data_file_open, errtext, stderr);
    if (err) {
	gui_errmsg(err, errtext);
	return;
    }	
    /* trash the practice files window that launched the query? */
    if (fwin) gtk_widget_destroy(fwin->w);    

    register_data(paths.datfile);
}

/* ........................................................... */

void verify_open_data (gpointer userdata)
     /* give user choice of not opening selected datafile,
	if there's already a datafile open and we're not
	in "expert" mode */
{
    if (data_file_open && expert[0] == 'f') { 
	edit_dialog ("gretl: open data", 
		     "Opening a new data file will automatically\n"
		     "close the current one.  Proceed?", NULL, 0,
		     "Yes", do_open_data, userdata, 
		     "No", NULL, NULL, 0, 0);
    } else 
	do_open_data(NULL, userdata);
}

/* ........................................................... */

void verify_open_session (gpointer userdata)
     /* give user choice of not opening session file,
	if there's already a datafile open and we're not
	in "expert" mode */
{
    if (data_file_open && expert[0] == 'f') { 
	edit_dialog ("gretl: open session", 
		     "Opening a new session file will automatically\n"
		     "close the current session.  Proceed?", NULL, 0,
		     "Yes", do_open_session, userdata, 
		     "No", NULL, NULL, 0, 0);
    } else 
	do_open_session(NULL, userdata);
}

/* ........................................................... */

static void set_data_from_filelist (gpointer data, guint i, 
				    GtkWidget *widget)
{
    strcpy(paths.datfile, datap[i]);
    verify_open_data(NULL);
}

/* ........................................................... */

static void set_session_from_filelist (gpointer data, guint i, 
				       GtkWidget *widget)
{
    strcpy(scriptfile, sessionp[i]);
    verify_open_session(NULL);
}

/* ........................................................... */

static void set_script_from_filelist (gpointer data, guint i, 
				      GtkWidget *widget)
{
    strcpy(scriptfile, scriptp[i]);
    do_open_script(NULL, NULL);
}

/* ........................................................... */

void save_session (char *fname) 
{
    int i, spos;
    char msg[MAXLEN], savedir[MAXLEN], fname2[MAXLEN];
    char session_base[MAXLEN], tmp[MAXLEN], grftmp[64];
    FILE *fp;
    print_t prn;

    spos = slashpos(fname);
    if (spos) 
	safecpy(savedir, fname, spos);
    else *savedir = 0;

    /* save commands, by dumping the command stack */
    if (haschar('.', fname) < 0)
	strcat(fname, ".gretl");
    if (dump_cmd_stack(fname)) return;

    get_base(session_base, fname, '.');

    /* get ready to save "session" */
    fp = fopen(fname, "a");
    if (fp == NULL) {
	sprintf(errtext, "Couldn't open session file %s", savefile);
	errbox(errtext);
	return;
    }
    fprintf(fp, "(* saved objects:\n");

    /* save session models */
    for (i=0; i<session.nmodels; i++) {
	fprintf(fp, "model %d \"%s\"\n", 
		(session.models[i])->ID, 
		(session.models[i])->name);
    }

    /* save session graphs */
    for (i=0; i<session.ngraphs; i++) {
	/* formulate save name for graph */
	strcpy(grftmp, (session.graphs[i])->name);
	sprintf(tmp, "%s%s", session_base, space_to_score(grftmp));
	/* does the constructed filename differ from the
	   current one? */
	if (strcmp((session.graphs[i])->fname, tmp)) {
	    if (copyfile((session.graphs[i])->fname, tmp)) {
		errbox("Couldn't copy graph file");
		continue;
	    } else {
		remove((session.graphs[i])->fname);
		strcpy((session.graphs[i])->fname, tmp);
	    }
	}
	fprintf(fp, "graph %d \"%s\" %s\n", 
		(session.graphs[i])->ID, 
		(session.graphs[i])->name, 
		(session.graphs[i])->fname);
    }

    fprintf(fp, "*)\n");
    fclose(fp);
    session_saved = 1;
    mkfilelist(2, fname);

    /* save output */
    switch_ext(fname2, fname, "txt");
    prn.fp = fopen(fname2, "w");
    if (prn.fp == NULL) {
	errbox("Couldn't open output file for writing");
	return;
    }
    execute_script(cmdfile, NULL, NULL, &prn, SESSION_EXEC); 
    fclose(prn.fp);

    if (savedir)
	sprintf(msg, "session saved to %s -\n", savedir);
    else 
	strcpy(msg, "session saved -\n");
    strcat(msg, "commands: ");
    strcat(msg, (spos)? fname + spos + 1 : fname);
    strcat(msg, "\n  output: ");
    switch_ext(fname2, fname, "txt");
    spos = slashpos(fname2);
    strcat(msg, (spos)? fname2 + spos + 1 : fname2);
    infobox(msg);

    return;
}

/* ......................................................... */

void helpfile_init (void)
{
    FILE *fp;
    char testline[MAXLEN];

    fp = fopen(paths.helpfile, "r");
    if (fp != NULL) { 
	while (fgets(testline, MAXLEN-1, fp)) gui_help_length++;
	fclose(fp);
    } else fprintf(stderr, "help file %s is not accessible\n", 
		   paths.helpfile);

    fp = fopen(paths.cmd_helpfile, "r");
    if (fp != NULL) { 
	while (fgets(testline, MAXLEN-1, fp)) script_help_length++;
	fclose(fp);
    } else fprintf(stderr, "help file %s is not accessible\n", 
		   paths.cmd_helpfile);

    help_length = gui_help_length;
}

/* ........................................................... */

void helpwin (gpointer data, guint script, GtkWidget *widget) 
{
    if (script) {
	help_length = script_help_length;
	view_file(paths.cmd_helpfile, 0, 0, 77, 400, 
		  "gretl: command syntax help", 
		  script_help_items, sizeof script_help_items);
    } else {
	help_length = gui_help_length;
	view_file(paths.helpfile, 0, 0, 77, 400, "gretl: help", 
		  help_items, sizeof help_items);
    }
}

/* ........................................................... */

static int help_index (const char *str)
{
    gchar helpline[84];
    gchar word[32];
    int found = 0, pos = 0;
    FILE *fp;

    fp = fopen(paths.helpfile, "r");
    while (!found) {
	if (fgets(helpline, 83, fp) == NULL) break;
	if (helpline[0] == '#') {
	    fgets(helpline, 83, fp);
	    sscanf(helpline, "%s", word);
	    if (strcmp(word, str) == 0) {
		found = 1;
	    } else pos++;
	}
	pos++;
    }
    fclose(fp);

    if (!found) return -1;
    return pos - 1;
}

/* ........................................................... */

void menu_find (gpointer data, guint db, GtkWidget *widget)
{
    if (db) 
	find_string_dialog(find_in_clist, cancel_find, data);
    else 
	find_string_dialog(find_in_help, cancel_find, data);
}

/* ........................................................... */

void datafile_find (GtkWidget *widget, gpointer data)
{
    find_string_dialog(find_in_clist, cancel_find, data);
}

/* ........................................................... */

void context_help (GtkWidget *widget, gpointer data)
{
    int help_code = GPOINTER_TO_INT(data);

    do_help(NULL, (guint) help_code, NULL);
}

/* ........................................................... */

void do_help (gpointer data, guint code, GtkWidget *widget) 
{
    int pos = 0;
    char cmdstr[12];

    if (code == MEANTEST2) code = MEANTEST;
    else if (code == MODEL_GENR) code = GENR;

    if (code == GR_PLOT || code == GR_XY || code == GNUPLOT)
	strcpy(cmdstr, "graphing");
    else if (code == GR_DUMMY)
	strcpy(cmdstr, "factorized");
    else if (code == ONLINE)
	strcpy(cmdstr, "online");
    else if (code == MARKERS)
	strcpy(cmdstr, "markers");
    else if (code == EXPORT)
	strcpy(cmdstr, "export");
    else if (code == SMPLBOOL || code == SMPLDUM)
	strcpy(cmdstr, "sampling");
    else if (code == PANEL) 
	strcpy(cmdstr, "panel");
    else if (code == COMPACT) 
	strcpy(cmdstr, "compact");
    else if (code < NC)
	strcpy(cmdstr, commands[code]);
    else
	pos = -1;

    if (!pos) pos = help_index(cmdstr);
    if (pos == -1) {
	errbox("Sorry, no help is available on this topic");
	return;
    }
    if (help_view == NULL) 
	helpwin(NULL, 0, NULL); 
    else
	gdk_window_raise(help_view->parent->window);
	
    gtk_adjustment_set_value(GTK_TEXT(help_view)->vadj, 
	(gfloat) pos * GTK_TEXT(help_view)->vadj->upper / help_length);
}

/* ........................................................... */

static void update_header (GtkWidget *widget, gpointer data)
{
    gchar *newhdr;
    FILE *fp;
    int err1, err2;

    fp = fopen(paths.hdrfile, "w");
    if (fp == NULL) {
	errbox("Can't open header file for writing");
	return;
    }
    newhdr = gtk_editable_get_chars (GTK_EDITABLE(data), 0, -1);
    err1 = fprintf(fp, "%s", newhdr);
    err2 = fclose(fp);
    if (err1 == 0 || err2) errbox("Failed to write header file");
    else infobox("Edited header saved OK");
    g_free(newhdr);
} 

/* ........................................................... */

static void close_console (void)
{    
    console_state(TRUE);
    console_view = NULL;
}

/* ........................................................... */

static void close_help (void)
{
    help_view = NULL;
}

/* .................................................................. */

void windata_init (windata_t *mydata)
{
    mydata->listbox = NULL;
    mydata->mbar = NULL;
    mydata->w = NULL;
    mydata->status = NULL;
    mydata->popup = NULL;
    mydata->ifac = NULL;
    mydata->data = NULL;
}

/* .................................................................. */

void free_windata (GtkWidget *w, gpointer data)
{
    windata_t *mydata = (windata_t *) data;

    if (mydata) {
	if (mydata->listbox) 
	    gtk_widget_destroy(GTK_WIDGET(mydata->listbox));
	if (mydata->mbar) 
	    gtk_widget_destroy(GTK_WIDGET(mydata->mbar));
	if (mydata->status) 
	    gtk_widget_destroy(GTK_WIDGET(mydata->status));
	if (mydata->ifac) 
	    gtk_object_unref(GTK_OBJECT(mydata->ifac));  
	if (mydata->popup) 
	    gtk_object_unref(GTK_OBJECT(mydata->popup));
	free(mydata);
	mydata = NULL;
    }
}

/* ........................................................... */

int view_buffer (print_t *prn, int hsize, int vsize, char *title, 
		 GtkItemFactoryEntry menu_items[], int msize) 
{
    GtkWidget *dialog, *close, *table;
    GdkFont *fixed_font;
    GtkWidget *vscrollbar; 
    windata_t *vwin;

    if ((vwin = mymalloc(sizeof *vwin)) == NULL)
	return 1;
    windata_init(vwin);

    fixed_font = gdk_font_load(fontspec); 
    hsize *= gdk_char_width(fixed_font, 'W');
    hsize += 48;

    dialog = gtk_dialog_new();
    gtk_widget_set_usize (dialog, hsize, vsize);
    gtk_window_set_title(GTK_WINDOW(dialog), title);
    gtk_container_border_width (GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), 5);
    gtk_container_border_width 
	(GTK_CONTAINER(GTK_DIALOG(dialog)->action_area), 5);
    gtk_box_set_spacing(GTK_BOX(GTK_DIALOG(dialog)->vbox), 5);
    gtk_box_set_homogeneous(GTK_BOX(GTK_DIALOG(dialog)->action_area), TRUE);
#ifndef G_OS_WIN32
    gtk_signal_connect_after(GTK_OBJECT(dialog), "realize", 
			     GTK_SIGNAL_FUNC(set_wm_icon), 
			     NULL);
#endif

    if (menu_items != NULL) {
	set_up_viewer_menu(dialog, vwin, menu_items, msize);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), 
			   vwin->mbar, FALSE, TRUE, 0);
	gtk_widget_show(vwin->mbar);
    }

    table = gtk_table_new(1, 2, FALSE);
    gtk_widget_set_usize(table, 500, 400);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), 
		       table, TRUE, TRUE, FALSE);

    vwin->w = gtk_text_new(NULL, NULL);

    /* special case: help viewer */
    if (strcmp(title, "gretl: help") == 0) {
	help_view = vwin->w;
	gtk_signal_connect(GTK_OBJECT(vwin->w), "destroy",
			   GTK_SIGNAL_FUNC(close_help), NULL);
    }

    gtk_text_set_editable(GTK_TEXT(vwin->w), FALSE);

    gtk_text_set_word_wrap(GTK_TEXT(vwin->w), TRUE);
    gtk_table_attach(GTK_TABLE(table), vwin->w, 0, 1, 0, 1,
		     GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND | 
		     GTK_SHRINK, 0, 0);
    gtk_widget_show(vwin->w);

    vscrollbar = gtk_vscrollbar_new(GTK_TEXT (vwin->w)->vadj);
    gtk_table_attach (GTK_TABLE (table), 
		      vscrollbar, 1, 2, 0, 1,
		      GTK_FILL, GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);
    gtk_widget_show (vscrollbar);

    gtk_widget_show(table);

    /* close button */
    close = gtk_button_new_with_label("Close");
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), 
		       close, FALSE, TRUE, 0);
    gtk_signal_connect(GTK_OBJECT(close), "clicked", 
		       GTK_SIGNAL_FUNC(delete_file_viewer), 
		       (gpointer) dialog);
    gtk_widget_show(close);

    /* insert and then free the text buffer */
    gtk_text_insert(GTK_TEXT(vwin->w), fixed_font, 
		    NULL, NULL, prn->buf, 
		    strlen(prn->buf));
    prnclose(prn);
    
    /* clean up when dialog is destroyed */
    gtk_signal_connect(GTK_OBJECT(dialog), "key_press_event", 
			   GTK_SIGNAL_FUNC(catch_key), dialog);

    gtk_signal_connect(GTK_OBJECT(dialog), "destroy", 
		       GTK_SIGNAL_FUNC(free_windata), vwin);

    gtk_widget_show(dialog);
    return 0;
}

/* ........................................................... */

int view_file (char *filename, int editable, int del_file, 
	       int hsize, int vsize, char *title, 
	       GtkItemFactoryEntry menu_items[], int msize) 
{
    GtkWidget *dialog, *close, *save = NULL, *table;
    GdkFont *fixed_font;
    GtkWidget *vscrollbar; 
    extern GdkColor red, blue;
    void *colptr = NULL, *nextcolor = NULL;
    char tempstr[MAXSTR], *fle;
    FILE *fd;
    windata_t *vwin;
    int console = 0;

    fd = fopen(filename, "r");
    if (fd == NULL) {
	sprintf(tempstr, "Can't open %s for reading", filename);
	errbox(tempstr);
	return 1;
    }

    if ((vwin = mymalloc(sizeof *vwin)) == NULL)
	return 1;
    windata_init(vwin);
    strcpy(vwin->fname, filename);

    fixed_font = gdk_font_load(fontspec); 
    hsize *= gdk_char_width(fixed_font, 'W');
    hsize += 48;

    if (strcmp(title, "gretl console") == 0) {
	console_state(FALSE);
	console = 1;
    }

    dialog = gtk_dialog_new();
    gtk_widget_set_usize (dialog, hsize, vsize);
    gtk_window_set_title(GTK_WINDOW(dialog), title);
    gtk_container_border_width (GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), 5);
    gtk_container_border_width 
        (GTK_CONTAINER(GTK_DIALOG(dialog)->action_area), 5);
    gtk_box_set_spacing(GTK_BOX(GTK_DIALOG(dialog)->vbox), 5);
    gtk_box_set_homogeneous(GTK_BOX(GTK_DIALOG(dialog)->action_area), TRUE);
#ifndef G_OS_WIN32
    gtk_signal_connect_after(GTK_OBJECT(dialog), "realize", 
			     GTK_SIGNAL_FUNC(set_wm_icon), 
			     NULL);
#endif

    if (menu_items != NULL) {
	set_up_viewer_menu(dialog, vwin, menu_items, msize);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), 
			   vwin->mbar, FALSE, TRUE, 0);
	gtk_widget_show(vwin->mbar);
    }

    table = gtk_table_new(1, 2, FALSE);
    gtk_widget_set_usize(table, 500, 400);
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), 
		       table, TRUE, TRUE, FALSE);

    vwin->w = gtk_text_new(NULL, NULL);

    /* special case: help viewer */
    if (strcmp(title, "gretl: help") == 0 ||
	strncmp(title, "gretl: command", 14) == 0) {
	help_view = vwin->w;
	gtk_signal_connect(GTK_OBJECT(vwin->w), "destroy",
			   GTK_SIGNAL_FUNC(close_help), NULL);
    }
    if (editable)
	gtk_text_set_editable(GTK_TEXT(vwin->w), TRUE);
    else 
	gtk_text_set_editable(GTK_TEXT(vwin->w), FALSE);

    /* special case: the gretl console */
    if (console) {
	gtk_signal_connect(GTK_OBJECT(vwin->w), "key_press_event",
			   (GtkSignalFunc) console_handler, NULL);
	gtk_signal_connect(GTK_OBJECT(vwin->w), "destroy",
			   GTK_SIGNAL_FUNC(close_console), NULL);
	console_view = vwin->w;
	console_dialog = dialog;
    }

    gtk_text_set_word_wrap(GTK_TEXT(vwin->w), TRUE);
    gtk_table_attach(GTK_TABLE(table), vwin->w, 0, 1, 0, 1,
		     GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND | 
		     GTK_SHRINK, 0, 0);
    gtk_widget_show(vwin->w);

    vscrollbar = gtk_vscrollbar_new (GTK_TEXT (vwin->w)->vadj);
    gtk_table_attach (GTK_TABLE (table), 
		      vscrollbar, 1, 2, 0, 1,
		      GTK_FILL, GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);
    gtk_widget_show (vscrollbar);

    gtk_widget_show(table);

    /* is the file to be deleted after viewing? */
    if (del_file) {
	if ((fle = mymalloc(strlen(filename) + 1)) == NULL)
	    return 1;
	strcpy(fle, filename);
    }

    /* add a "save" button for editing hdr file */
    if (strncmp(title, "gretl: edit", 11) == 0)  {
	save = gtk_button_new_with_label("Save");
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), 
			   save, FALSE, TRUE, 0);
	gtk_signal_connect(GTK_OBJECT(save), "clicked", 
			   GTK_SIGNAL_FUNC(update_header), 
			   (gpointer) vwin->w); 
	gtk_signal_connect_after(GTK_OBJECT(save), "clicked", 
			   GTK_SIGNAL_FUNC(delete_file_viewer), 
			   (gpointer) dialog);
	gtk_widget_show(save);
    }

    /* close button for all uses */
    close = gtk_button_new_with_label("Close");
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), 
		       close, FALSE, TRUE, 0);
    gtk_signal_connect(GTK_OBJECT(close), "clicked", 
		       GTK_SIGNAL_FUNC(delete_file_viewer), 
		       (gpointer) dialog);
    gtk_widget_show(close);

    /* insert the file text */
    memset(tempstr, 0, sizeof tempstr);
    while (fgets(tempstr, sizeof tempstr - 1, fd)) {
	if (tempstr[0] == '?') 
	    colptr = (console)? &red : &blue;
	if (tempstr[0] == '#') {
	    tempstr[0] = ' ';
	    nextcolor = &red;
	} else
	    nextcolor = NULL;
	gtk_text_insert(GTK_TEXT(vwin->w), fixed_font, 
			colptr, NULL, tempstr, 
			strlen(tempstr));
	colptr = nextcolor;
	memset(tempstr, 0, sizeof tempstr);
    }
    fclose(fd);

    /* clean up when dialog is destroyed */
    if (del_file) {
	gtk_signal_connect(GTK_OBJECT(dialog), "destroy", 
			   GTK_SIGNAL_FUNC(delete_file), (gpointer) fle);
    }
    if (!editable) {
	gtk_signal_connect(GTK_OBJECT(dialog), "key_press_event", 
			   GTK_SIGNAL_FUNC(catch_key), dialog);
    }
    gtk_signal_connect(GTK_OBJECT(dialog), "destroy", 
		       GTK_SIGNAL_FUNC(free_windata), vwin);

    gtk_widget_show(dialog);
    return 0;
}

/* ........................................................... */

void flip (GtkItemFactory *ifac, char *path, gboolean s)
{
    if (ifac != NULL)
	gtk_widget_set_sensitive(gtk_item_factory_get_item
				 (ifac, path), s);
}

/* ........................................................... */

static void model_panel_menu_state (GtkItemFactory *ifac, gboolean s)
{
    flip(ifac, "/Tests/panel diagnostics", s);
}

/* ........................................................... */

static void model_menu_state (GtkItemFactory *ifac, gboolean s)
{
    flip(ifac, "/Tests/non-linearity (squares)", s);
    flip(ifac, "/Tests/non-linearity (logs)", s);
    flip(ifac, "/Tests/autocorrelation", s);
    flip(ifac, "/Tests/heteroskedasticity", s);
    flip(ifac, "/Tests/Chow test", s);
    flip(ifac, "/Tests/CUSUM test", s);
    flip(ifac, "/Tests/ARCH", s);
    flip(ifac, "/Tests/normality of residual", s);
    flip(ifac, "/Graphs", s);
    flip(ifac, "/Model data/Display actual, fitted, residual", s);
    flip(ifac, "/Model data/Forecasts with standard errors", s);
    flip(ifac, "/Model data/Add to data set/residuals", s);
    flip(ifac, "/Model data/Add to data set/error sum of squares", s);
    flip(ifac, "/Model data/Add to data set/standard error of residuals", s);
    flip(ifac, "/Model data/Add to data set/R-squared", s);
}

/* ........................................................... */

static void lmmenu_state (GtkItemFactory *ifac, gboolean s)
{
    flip(ifac, "/Tests/non-linearity (squares)", s);
    flip(ifac, "/Tests/non-linearity (logs)", s);
    flip(ifac, "/Tests/autocorrelation", s);
    flip(ifac, "/Tests/heteroskedasticity", s);
    flip(ifac, "/Tests/Chow test", s);
    flip(ifac, "/Tests/CUSUM test", s);
    flip(ifac, "/Tests/ARCH", s);
}

/* ........................................................... */

static void latex_menu_state (GtkItemFactory *ifac, gboolean s)
{
    flip(ifac, "/LaTeX", s);
}

/* ........................................................... */

static void model_save_state (GtkItemFactory *ifac, gboolean s)
{
    flip(ifac, "/File/Save to session as icon...", s);
    flip(ifac, "/File/Save as icon and close", s);
}

/* ........................................................... */

static void set_up_viewer_menu (GtkWidget *window, windata_t *vwin, 
				GtkItemFactoryEntry items[], int msize)
{
    GtkAccelGroup *accel;
    gint n_items = msize / sizeof items[0];

    accel = gtk_accel_group_new();
    vwin->ifac = gtk_item_factory_new(GTK_TYPE_MENU_BAR, "<main>", 
				      accel);
    gtk_item_factory_create_items(vwin->ifac, n_items, items, vwin);
    vwin->mbar = gtk_item_factory_get_widget(vwin->ifac, "<main>");
    gtk_accel_group_attach(accel, GTK_OBJECT (window));

    if (vwin->data) {  /* FIXME? what if "data" is not ptr to model? */
	MODEL *pmod = (MODEL *) vwin->data;

	if (pmod->ci == POOLED) 
	    model_panel_menu_state(vwin->ifac, TRUE);
	else model_panel_menu_state(vwin->ifac, FALSE);
	if (pmod->ci != OLS && pmod->ci != POOLED) { 
	    lmmenu_state(vwin->ifac, FALSE);
	    latex_menu_state(vwin->ifac, FALSE);
	}
	if (pmod->ci == LOGIT || pmod->ci == PROBIT) {
	    model_menu_state(vwin->ifac, FALSE);
	}
	if (pmod->name)
	    model_save_state(vwin->ifac, FALSE);
    }
}

/* .................................................................. */

static void add_vars_to_plot_menu (windata_t *vwin)
{
    int i, j;
    GtkItemFactoryEntry varitem;
    gchar *mpath[] = {"/Graphs/residual plot", 
		      "/Graphs/fitted, actual plot"};
    MODEL *pmod = vwin->data;

    varitem.path = NULL;

    for (i=0; i<2; i++) {
	varitem.path = mymalloc(64);
	varitem.accelerator = NULL;
	varitem.callback_action = 0; 
	varitem.item_type = NULL;
	if (dataset_is_time_series(datainfo))
	    sprintf(varitem.path, "%s/against time", mpath[i]);
	else
	    sprintf(varitem.path, "%s/by observation number", mpath[i]);
	if (i == 0)
	    varitem.callback = resid_plot; 
	else
	    varitem.callback = fit_actual_plot;
	gtk_item_factory_create_item(vwin->ifac, &varitem, vwin, 1);

	/* put the indep vars on the menu list */
	for (j=2; j<pmod->list[0]; j++) {
	    if (pmod->list[j] == 0) continue;
	    if (varitem.path == NULL)
		varitem.path = mymalloc(64);
	    varitem.accelerator = NULL;
	    varitem.callback_action = pmod->list[j]; 
	    varitem.item_type = NULL;
	    sprintf(varitem.path, "%s/against %s", mpath[i], 
		    datainfo->varname[pmod->list[j]]);
	    if (i == 0)
		varitem.callback = resid_plot; 
	    else
		varitem.callback = fit_actual_plot;
	    gtk_item_factory_create_item(vwin->ifac, &varitem, vwin, 1);
	}
    }
    free(varitem.path);
}

/* ........................................................... */

static void check_model_menu (GtkWidget *w, GdkEventButton *eb, 
			      gpointer data)
{
    windata_t *mwin = (windata_t *) data;
    MODEL *pmod = mwin->data;
    extern int quiet_sample_check (MODEL *pmod);
    int s, ok = 1;

    if (quiet_sample_check(pmod)) ok = 0;
    s = GTK_WIDGET_IS_SENSITIVE
	(gtk_item_factory_get_item(mwin->ifac, "/Tests/omit variables"));
    if ((s && ok) || (!s && !ok)) return;
    s = !s;

    flip(mwin->ifac, "/Tests/omit variables", s);
    flip(mwin->ifac, "/Tests/add variables", s);
    flip(mwin->ifac, "/Tests/non-linearity (squares)", s);
    flip(mwin->ifac, "/Tests/non-linearity (logs)", s);
    flip(mwin->ifac, "/Tests/autocorrelation", s);
    flip(mwin->ifac, "/Tests/heteroskedasticity", s);
    flip(mwin->ifac, "/Tests/Chow test", s);
    flip(mwin->ifac, "/Tests/CUSUM test", s);
    flip(mwin->ifac, "/Tests/ARCH", s);
    flip(mwin->ifac, "/Graphs", s);
    flip(mwin->ifac, "/Model data/Display actual, fitted, residual", s);
    flip(mwin->ifac, "/Model data/Forecasts with standard errors", s);
    flip(mwin->ifac, "/Model data/Confidence intervals for coefficients", s);
    flip(mwin->ifac, "/Model data/Add to data set/fitted values", s);
    flip(mwin->ifac, "/Model data/Add to data set/residuals", s);
    flip(mwin->ifac, "/Model data/Add to data set/squared residuals", s);
    flip(mwin->ifac, "/Model data/Define new variable...", s);
}

/* ........................................................... */

int view_model (print_t *prn, MODEL *pmod, int hsize, int vsize, 
		char *title) 
{
    windata_t *vwin;
    GtkWidget *dialog, *close, *table, *vscrollbar;
    GdkFont *fixed_font;

    if ((vwin = mymalloc(sizeof *vwin)) == NULL) return 1;
    windata_init(vwin);

    fixed_font = gdk_font_load(fontspec);
    hsize *= gdk_char_width (fixed_font, 'W');
    hsize += 48;

    vwin->data = pmod;
    dialog = gtk_dialog_new();
    gtk_widget_set_usize (dialog, hsize, vsize);
    gtk_window_set_title(GTK_WINDOW(dialog), title);
    gtk_container_border_width (GTK_CONTAINER 
				(GTK_DIALOG(dialog)->vbox), 5);
    gtk_container_border_width 
	(GTK_CONTAINER(GTK_DIALOG(dialog)->action_area), 5);
    gtk_box_set_spacing(GTK_BOX(GTK_DIALOG(dialog)->vbox), 5);
    gtk_box_set_homogeneous(GTK_BOX 
			    (GTK_DIALOG(dialog)->action_area), TRUE);
#ifndef G_OS_WIN32
    gtk_signal_connect_after(GTK_OBJECT(dialog), "realize", 
			     GTK_SIGNAL_FUNC(set_wm_icon), 
			     NULL);
#endif

    set_up_viewer_menu(dialog, vwin, model_items, sizeof model_items);
    /* add menu of indep vars, against which to plot resid */
    add_vars_to_plot_menu(vwin);
    gtk_signal_connect(GTK_OBJECT(vwin->mbar), "button_press_event", 
		       GTK_SIGNAL_FUNC(check_model_menu), vwin);

    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), 
		       vwin->mbar, FALSE, TRUE, 0);
    gtk_widget_show(vwin->mbar);

    table = gtk_table_new(1, 2, FALSE);
    gtk_widget_set_usize(table, hsize, vsize); 
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), 
		       table, TRUE, TRUE, FALSE);

    vwin->w = gtk_text_new(NULL, NULL);
    gtk_text_set_editable(GTK_TEXT(vwin->w), FALSE);
    gtk_text_set_word_wrap(GTK_TEXT(vwin->w), TRUE);
    gtk_table_attach(GTK_TABLE(table), vwin->w, 0, 1, 0, 1,
		     GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND | 
		     GTK_SHRINK, 0, 0);
    gtk_widget_show(vwin->w);

    vscrollbar = gtk_vscrollbar_new (GTK_TEXT(vwin->w)->vadj);
    gtk_table_attach (GTK_TABLE (table), 
		      vscrollbar, 1, 2, 0, 1,
		      GTK_FILL, GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);
    gtk_widget_show (vscrollbar);
    gtk_widget_show(table);

    /* close button */
    close = gtk_button_new_with_label("Close");
    gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area), 
		       close, FALSE, TRUE, 0);
    gtk_signal_connect(GTK_OBJECT(close), "clicked", 
		       GTK_SIGNAL_FUNC(delete_file_viewer), 
		       (gpointer) dialog);
    gtk_widget_show(close);

    /* insert and then free the model buffer */
    gtk_text_insert(GTK_TEXT(vwin->w), fixed_font, 
		    NULL, NULL, prn->buf, 
		    strlen(prn->buf));
/*      fprintf(stderr, "view_model: freeing model buffer\n"); */
    prnclose(prn);

    copylist(&default_list, pmod->list);

    /* attach shortcuts */
    gtk_object_set_data(GTK_OBJECT(dialog), "ddata", vwin);
    gtk_signal_connect(GTK_OBJECT(dialog), "key_press_event", 
		       GTK_SIGNAL_FUNC(catch_key), 
		       dialog);
    gtk_signal_connect(GTK_OBJECT(dialog), "key_press_event", 
		       GTK_SIGNAL_FUNC(catch_key), 
		       dialog);

    /* clean up when dialog is destroyed */
    gtk_signal_connect(GTK_OBJECT(dialog), "destroy", 
		       GTK_SIGNAL_FUNC(delete_model), 
		       vwin->data);
    gtk_signal_connect(GTK_OBJECT(dialog), "destroy", 
		       GTK_SIGNAL_FUNC(free_windata), 
		       vwin);

    gtk_widget_show(dialog);
    return 0;
}

/* ......................................................... */

void setup_column (GtkWidget *listbox, int column, int width) 
{
    if (width == 0) 
	gtk_clist_set_column_auto_resize (GTK_CLIST (listbox), column, TRUE);
    else if (width == -1) 
	gtk_clist_set_column_visibility (GTK_CLIST (listbox), column, FALSE);
    else 
	gtk_clist_set_column_width (GTK_CLIST (listbox), column, width);
}

/* ......................................................... */

void yes_no_dialog (char *diagtxt, char *infotxt, int erase, 
		    int num, ...) 
{
   GtkWidget *text, *tempwid, *dialog;
   typedef void (*func)();
   dialog_t *d;
   func myfunc;
   va_list argp;
   char *tempstr;
   GList *list;
   void *ptr;
   int i;
   
   dialog = gtk_dialog_new ();
   gtk_grab_add (dialog);
   gtk_window_set_title (GTK_WINDOW (dialog), diagtxt);
   gtk_window_set_policy (GTK_WINDOW (dialog), FALSE, FALSE, FALSE);
   gtk_container_border_width 
       (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), 10);
   gtk_container_border_width 
       (GTK_CONTAINER (GTK_DIALOG (dialog)->action_area), 5);
   gtk_box_set_spacing (GTK_BOX (GTK_DIALOG (dialog)->vbox), 5);
   gtk_box_set_spacing (GTK_BOX (GTK_DIALOG (dialog)->action_area), 15);
   gtk_box_set_homogeneous (GTK_BOX (GTK_DIALOG (dialog)->action_area), TRUE);
   gtk_window_set_position (GTK_WINDOW (dialog), GTK_WIN_POS_MOUSE);
   
   text = gtk_label_new (infotxt);
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), text, 
		       TRUE, TRUE, FALSE);
   gtk_widget_show(text);

   list = NULL;
   ptr = g_malloc(1);
   g_list_append(list, ptr);
   va_start (argp, num);
   for (i=0; i<num; i++) {
      tempstr = va_arg (argp, char *);
      myfunc = va_arg (argp, func);
      ptr = va_arg (argp, void *);
      d = mymalloc(sizeof *d);
      g_list_append(list, d);
      d->dialog = dialog;
      d->data = ptr;
      d->all_buttons = list;
      tempwid = gtk_button_new_with_label(tempstr);
      gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), 
			  tempwid, TRUE, TRUE, TRUE);
      if (myfunc) {
         gtk_signal_connect (GTK_OBJECT (tempwid), "clicked", 
			     GTK_SIGNAL_FUNC (myfunc), d);
      }
      if (erase || !myfunc) {
         gtk_signal_connect_object (GTK_OBJECT (tempwid), 
				    "clicked", 
				    GTK_SIGNAL_FUNC (gtk_widget_destroy), 
				    (gpointer) d->dialog);
      }
      gtk_widget_show (tempwid);
   }
            
   gtk_widget_show (dialog);
}

/* ........................................................... */

static void msgbox (const char *msg, int err) 
{
    GtkWidget *w, *label, *button, *table;
    char labeltext[MAXLEN];

    if (err)
	sprintf(labeltext, "Error:\n%s\n", msg);
    else
	sprintf(labeltext, "Info:\n%s\n", msg);
    w = gtk_window_new(GTK_WINDOW_DIALOG);
    gtk_container_border_width(GTK_CONTAINER(w), 5);
    gtk_window_position (GTK_WINDOW(w), GTK_WIN_POS_MOUSE);
    gtk_window_set_title (GTK_WINDOW (w), (err)? "gretl error" : 
			  "gretl info");  
  
    table = gtk_table_new(2, 3, FALSE);
    gtk_container_add(GTK_CONTAINER(w), table);
  
    label = gtk_label_new(labeltext);
    gtk_table_attach_defaults(GTK_TABLE(table), label, 0, 3, 0, 1);

    if (err)
	button = gtk_button_new_with_label("Close");
    else
	button = gtk_button_new_with_label("OK");
    gtk_table_attach_defaults(GTK_TABLE(table), button, 1, 2, 1, 2);
  
    gtk_signal_connect(GTK_OBJECT(button), "clicked",
		       GTK_SIGNAL_FUNC(delete_widget), w);
    gtk_widget_show(button);
    gtk_widget_show(label);
    gtk_widget_show(table);
    gtk_widget_show(w);  
}

/* ........................................................... */

void errbox (const char *msg) 
{
    msgbox(msg, 1);
}

/* ........................................................... */

void infobox (const char *msg) 
{
    msgbox(msg, 0);
}

/* ........................................................... */

int validate_varname (const char *varname)
{
    int i, n = strlen(varname);
    char errtext[MAXLEN], namebit[9];
    
    if (n > 8) {
	safecpy(namebit, varname, 8);
	sprintf(errtext, "Variable name %s... is too long\n"
	       "(the max is 8 characters)", namebit);
	errbox(errtext);
	return 1;
    }
    if (!(isalpha(varname[0]))) {
	sprintf(errtext, "First char of name ('%c') is bad\n"
	       "(first must be alphabetical)", varname[0]);
	errbox(errtext);
	return 1;
    }
    for (i=1; i<n; i++) {
	if (!(isalpha(varname[i]))  
	    && !(isdigit(varname[i]))
	    && varname[i] != '_') {
	    sprintf(errtext, "Name contains an illegal char (in place %d)\n"
		    "Use only letters, digits and underscore", i + 1);
	    errbox(errtext);
	    return 1;
	}
    }
    return 0;
}	

/* .................................................................. */

void options_dialog (gpointer data) 
{
    GtkWidget *tempwid, *dialog, *notebook;

    dialog = gtk_dialog_new ();
    gtk_window_set_title (GTK_WINDOW (dialog), "gretl: options");
    gtk_container_border_width 
	(GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), 10);
    gtk_container_border_width 
	(GTK_CONTAINER (GTK_DIALOG (dialog)->action_area), 5);
    gtk_box_set_spacing (GTK_BOX (GTK_DIALOG (dialog)->vbox), 2);
    gtk_box_set_spacing (GTK_BOX (GTK_DIALOG (dialog)->action_area), 15);
    gtk_box_set_homogeneous (GTK_BOX (GTK_DIALOG (dialog)->action_area), TRUE);
    gtk_window_set_position (GTK_WINDOW (dialog), GTK_WIN_POS_MOUSE);
    gtk_signal_connect_object 
	(GTK_OBJECT (dialog), "delete_event", GTK_SIGNAL_FUNC 
	 (gtk_widget_destroy), GTK_OBJECT (dialog));
   
    notebook = gtk_notebook_new ();
    gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), 
			notebook, TRUE, TRUE, 0);
    gtk_widget_show (notebook);

    make_prefs_tab (notebook, 1);
    make_prefs_tab (notebook, 2);
    make_prefs_tab (notebook, 3);
   
    tempwid = gtk_button_new_with_label ("OK");
    GTK_WIDGET_SET_FLAGS (tempwid, GTK_CAN_DEFAULT);
    gtk_box_pack_start (GTK_BOX (GTK_DIALOG 
				 (dialog)->action_area), 
			tempwid, TRUE, TRUE, 0);
    gtk_signal_connect (GTK_OBJECT (tempwid), "clicked", 
			GTK_SIGNAL_FUNC (apply_changes), NULL);
    gtk_signal_connect_object (GTK_OBJECT (tempwid), "clicked", 
			       GTK_SIGNAL_FUNC (gtk_widget_destroy), 
			       GTK_OBJECT (dialog));
    gtk_widget_show (tempwid);

    tempwid = gtk_button_new_with_label ("  Cancel  ");
    GTK_WIDGET_SET_FLAGS (tempwid, GTK_CAN_DEFAULT);
    gtk_box_pack_start (GTK_BOX (GTK_DIALOG 
				 (dialog)->action_area), 
			tempwid, TRUE, TRUE, 0);
    gtk_signal_connect_object (GTK_OBJECT (tempwid), "clicked", 
			       GTK_SIGNAL_FUNC (gtk_widget_destroy), 
			       GTK_OBJECT (dialog));
    gtk_widget_show (tempwid);

    tempwid = gtk_button_new_with_label ("Apply");
    GTK_WIDGET_SET_FLAGS (tempwid, GTK_CAN_DEFAULT);
    gtk_box_pack_start (GTK_BOX (GTK_DIALOG 
				 (dialog)->action_area), 
			tempwid, TRUE, TRUE, 0);
    gtk_signal_connect (GTK_OBJECT (tempwid), "clicked", 
			GTK_SIGNAL_FUNC (apply_changes), NULL);
    gtk_widget_grab_default (tempwid);
    gtk_widget_show (tempwid);

    gtk_widget_show (dialog);
}

/* .................................................................. */

static void make_prefs_tab (GtkWidget *notebook, int tab) 
{
    GtkWidget *tempwid, *box, *inttbl, *chartbl;
    int num, tbl_len, tbl_num, tbl_col;
   
    box = gtk_vbox_new (FALSE, 0);
    gtk_container_border_width (GTK_CONTAINER (box), 10);
    gtk_widget_show (box);

    if (tab == 1)
	tempwid = gtk_label_new ("General");
    else if (tab == 2)
	tempwid = gtk_label_new ("Databases");
    else if (tab == 3)
	tempwid = gtk_label_new ("Toolbar");
    gtk_widget_show (tempwid);
    gtk_notebook_append_page (GTK_NOTEBOOK (notebook), box, tempwid);   

    tbl_len = 1;
    chartbl = gtk_table_new (tbl_len, 2, FALSE);
    gtk_table_set_row_spacings (GTK_TABLE (chartbl), 5);
    gtk_table_set_col_spacings (GTK_TABLE (chartbl), 5);
    gtk_box_pack_start (GTK_BOX (box), chartbl, FALSE, FALSE, 0);
    gtk_widget_show (chartbl);
   
    tbl_num = tbl_col = 0;
    inttbl = gtk_table_new (1, 2, FALSE);
    gtk_table_set_row_spacings (GTK_TABLE (inttbl), 2);
    gtk_table_set_col_spacings (GTK_TABLE (inttbl), 5);
    gtk_box_pack_start (GTK_BOX (box), inttbl, FALSE, FALSE, 0);
    gtk_widget_show (inttbl);

    num = 0;
    while (rc_vars[num].var != NULL) {
	if (rc_vars[num].tab == tab) {
	    if (rc_vars[num].type == 'I') {
		tempwid = gtk_check_button_new_with_label 
		    (rc_vars[num].description);
		gtk_table_attach_defaults 
		    (GTK_TABLE (inttbl), tempwid, tbl_col, tbl_col + 1, 
		     tbl_num, tbl_num + 1);
		if (strcmp(rc_vars[num].var, "true") == 0)
		    gtk_toggle_button_set_active 
			(GTK_TOGGLE_BUTTON (tempwid), TRUE);
		else
		    gtk_toggle_button_set_active 
			(GTK_TOGGLE_BUTTON (tempwid), FALSE);
		gtk_widget_show (tempwid);
		rc_vars[num].widget = tempwid;
		tbl_col++;
		if (tbl_col == 2) {
		    tbl_col = 0;
		    tbl_num++;
		    gtk_table_resize (GTK_TABLE (inttbl), tbl_num + 1, 2);
		}
	    } else {
		tbl_len++;
		gtk_table_resize (GTK_TABLE (chartbl), tbl_len, 2);
		tempwid = gtk_label_new (rc_vars[num].description);
		gtk_misc_set_alignment (GTK_MISC (tempwid), 1, 0.5);
		gtk_table_attach_defaults (GTK_TABLE (chartbl), 
					   tempwid, 0, 1, tbl_len-1, tbl_len);
		gtk_widget_show (tempwid);

		tempwid = gtk_entry_new ();
		gtk_table_attach_defaults (GTK_TABLE (chartbl), 
					   tempwid, 1, 2, tbl_len-1, tbl_len);
		gtk_entry_set_text (GTK_ENTRY (tempwid), rc_vars[num].var);
		gtk_widget_show (tempwid);
		rc_vars[num].widget = tempwid;
	    } 
	}
	num++;
    }
}

/* .................................................................. */

static void apply_changes (GtkWidget *widget, gpointer data) 
{
    gchar *tempstr;
    extern void show_toolbar (void);
    int i = 0;

    while (rc_vars[i].var != NULL) {
	if (rc_vars[i].widget != NULL) {
	    if (rc_vars[i].type == 'I') {
		if (GTK_TOGGLE_BUTTON(rc_vars[i].widget)->active)
		    strcpy(rc_vars[i].var, "true");
		else strcpy(rc_vars[i].var, "false");
	    } 
	    if (rc_vars[i].type == 'C') {
		tempstr = gtk_entry_get_text
		    (GTK_ENTRY(rc_vars[i].widget));
		if (tempstr != NULL && strlen(tempstr)) 
		    strncpy(rc_vars[i].var, tempstr, MAXSTR - 1);
	    }
	}
	i++;
    }
    write_rc_file();
    if (toolbar_box == NULL && want_toolbar[0] == 't')
	show_toolbar();
    else if (toolbar_box != NULL && want_toolbar[0] == 'f') {
	gtk_widget_destroy(toolbar_box);
	toolbar_box = NULL;
    }
}

/* .................................................................. */

void write_rc_file (void) 
{
    FILE *rc;
    int i;

    rc = fopen(rcfile, "w");
    if (rc == NULL) {
	errbox("Couldn't open config file for writing");
	return;
    }
    fprintf(rc, "# config file written by gretl: do not edit\n");
    i = 0;
    while (rc_vars[i].var != NULL) {
	fprintf(rc, "# %s\n", rc_vars[i].description);
	fprintf(rc, "%s = %s\n", rc_vars[i].key, rc_vars[i].var);
	i++;
    }
    printfilelist(1, rc); /* data files */
    printfilelist(2, rc); /* session files */
    printfilelist(3, rc); /* script files */
    fclose(rc);
}

/* .................................................................. */

static void read_rc_file (void) 
{
    FILE *rc;
    int i, j, numvars = sizeof rc_vars / sizeof *rc_vars;
    char line[MAXLEN], key[32], linevar[MAXLEN];
    int gotrecent = 0;

    if ((rc = fopen(rcfile, "r")) == NULL) return;

    i = 0;
    while (rc_vars[i].var != NULL) {
	if (fgets(line, MAXLEN, rc) == NULL) 
	    break;
	if (line[0] == '#') 
	    continue;
	if (!strncmp(line, "recent ", 7)) {
	    gotrecent = 1;
	    break;
	}
	if (sscanf(line, "%s", key) == 1) {
	    strcpy(linevar, line + strlen(key) + 3); 
	    chopstr(linevar); 
	    for (j=0; j<numvars; j++) {
		if (!strcmp(key, rc_vars[j].key))
		    strcpy(rc_vars[j].var, linevar);
	    }
	}
	i++;
    }

    if (expert[0] == '1') strcpy(expert, "true");
    else if (expert[0] == '0') strcpy(expert, "false");
    if (want_toolbar[0] == '1') strcpy(want_toolbar, "true");
    if (want_toolbar[0] == '0') strcpy(want_toolbar, "false");

    /* get lists of recently opened files */
    for (i=0; i<MAXRECENT; i++) { 
	datalist[i][0] = 0;
	sessionlist[i][0] = 0;
	scriptlist[i][0] = 0;
    }
    if (gotrecent || (fgets(line, MAXLEN, rc) != NULL && 
	strncmp(line, "recent data files:", 18) == 0)) {
	i = 0;
	while (fgets(line, MAXLEN, rc) && i<MAXRECENT) {
	    if (strncmp(line, "recent session files:", 21) == 0)
		break;
	    chopstr(line);
	    if (strlen(line)) 
		strcpy(datalist[i++], line);
	}
    }
    if (strncmp(line, "recent session files:", 21) == 0) {
	i = 0;
	while (fgets(line, MAXLEN, rc) && i<MAXRECENT) {
	    if (strncmp(line, "recent script files:", 20) == 0)
		break;
	    chopstr(line);
	    if (strlen(line)) 
		strcpy(sessionlist[i++], line);
	}
    }
    if (strncmp(line, "recent script files:", 20) == 0) {
	i = 0;
	while (fgets(line, MAXLEN, rc) && i<MAXRECENT) {
	    chopstr(line);
	    if (strlen(line)) 
		strcpy(scriptlist[i++], line);
	}
    }
#ifndef G_OS_WIN32
    set_paths(&paths, 1, 1);
#endif
    fclose(rc);
}

/* .................................................................. */

static void font_selection_ok (GtkWidget *w, GtkFontSelectionDialog *fs)
{
    gchar *fstring = gtk_font_selection_dialog_get_font_name(fs);

    if (strlen(fstring))
	strcpy(fontspec, fstring);
    write_rc_file();
    g_free(fstring);
    gtk_widget_destroy(GTK_WIDGET (fs));
}

/* .................................................................. */

void font_selector (void)
{
    static GtkWidget *fontsel = NULL;
    gchar *spacings[] = { "c", "m", NULL };

    if (!fontsel) {
	fontsel = gtk_font_selection_dialog_new 
	    ("Font for gretl output windows");
	gtk_window_set_position (GTK_WINDOW (fontsel), GTK_WIN_POS_MOUSE);
	gtk_font_selection_dialog_set_filter 
	    (GTK_FONT_SELECTION_DIALOG (fontsel),
                                       GTK_FONT_FILTER_BASE, GTK_FONT_ALL,
                                       NULL, NULL, NULL, NULL, spacings, NULL);
	gtk_font_selection_dialog_set_font_name 
	    (GTK_FONT_SELECTION_DIALOG (fontsel), fontspec);

	gtk_signal_connect (GTK_OBJECT(fontsel), "destroy",
			    GTK_SIGNAL_FUNC(gtk_widget_destroyed),
			    &fontsel);

	gtk_signal_connect (GTK_OBJECT 
			    (GTK_FONT_SELECTION_DIALOG 
			     (fontsel)->ok_button),
			    "clicked", GTK_SIGNAL_FUNC(font_selection_ok),
			    GTK_FONT_SELECTION_DIALOG (fontsel));

	gtk_signal_connect_object (GTK_OBJECT 
				   (GTK_FONT_SELECTION_DIALOG 
				    (fontsel)->cancel_button),
				   "clicked", 
				   GTK_SIGNAL_FUNC(gtk_widget_destroy),
				   GTK_OBJECT (fontsel));
    }
    if (!GTK_WIDGET_VISIBLE (fontsel)) gtk_widget_show (fontsel);
    else gtk_widget_destroy (fontsel);
}

/* .................................................................. */

static void close_find_dialog (GtkWidget *widget, gpointer data)
{
    gtk_widget_destroy (widget);
    find_window = NULL;
}

/* .................................................................. */

static void find_in_help (GtkWidget *widget, gpointer data)
{
    int nIndex = 0, i, linecount = 0;
    char *haystack;   

    haystack = gtk_editable_get_chars(GTK_EDITABLE(help_view), 0,
	gtk_text_get_length(GTK_TEXT(help_view)));

    if (needle) g_free(needle);

    needle = gtk_editable_get_chars(GTK_EDITABLE (find_entry), 0, -1);
    nIndex = GTK_EDITABLE (help_view)->selection_end_pos;

    nIndex = look_for_string(haystack, needle, nIndex);

    if (nIndex >= 0) {
	gtk_text_freeze(GTK_TEXT(help_view));
        gtk_text_set_point (GTK_TEXT (help_view), nIndex);
        gtk_text_insert (GTK_TEXT (help_view), NULL, NULL, NULL, " ", 1);
        gtk_text_backward_delete (GTK_TEXT (help_view), 1);
	gtk_text_thaw(GTK_TEXT(help_view));
        gtk_editable_select_region (GTK_EDITABLE (help_view), 
				    nIndex, nIndex + strlen (needle));
	for (i=0; i<nIndex; i++) 
	    if (haystack[i] == '\n') linecount++;
	gtk_adjustment_set_value(GTK_TEXT(help_view)->vadj, 
				 (gfloat) (linecount - 2) *
				 GTK_TEXT(help_view)->vadj->upper / help_length);
	find_window = NULL;
    } else infobox("String was not found.");

    g_free(haystack);
}

/* .................................................................. */

static void find_in_clist (GtkWidget *w, gpointer data)
{
    int start, found = 0, n, i;
    gchar *tmp; 
    char haystack[MAXLEN];
    windata_t *dbdat;

    dbdat = (windata_t *) gtk_object_get_data(GTK_OBJECT(data), "windat");

    if (needle) g_free(needle);
    needle = gtk_editable_get_chars(GTK_EDITABLE (find_entry), 0, -1);
    lower(needle);

    start = dbdat->active_var + 1;
    n = GTK_CLIST(dbdat->listbox)->rows;

    for (i=start; i<n; i++) {  
	gtk_clist_get_text(GTK_CLIST(dbdat->listbox), i, 1, &tmp);
	strcpy(haystack, tmp);
	lower(haystack);
	found = look_for_string(haystack, needle, 0);
	if (found >= 0) break;
    }
    if (found >= 0) {
	gtk_clist_moveto(GTK_CLIST(dbdat->listbox), i, 0, 0, .1);
	gtk_clist_select_row(GTK_CLIST(dbdat->listbox), i, 0);
	dbdat->active_var = i;
	find_window = NULL;    
    } else {
	gtk_clist_select_row(GTK_CLIST(dbdat->listbox), 0, 0);
	dbdat->active_var = 0;
	infobox("String was not found.");
    }
}

/* .................................................................. */

static int look_for_string (char *haystack, char *needle, int start)
{
    int pos;
    int HaystackLength = strlen(haystack);
    int NeedleLength = strlen(needle);

    for (pos = start; pos < HaystackLength; pos++) {
        if (strncmp(&haystack[pos], needle, NeedleLength) == 0) 
             return pos;
    }
    return -1;
}

/* .................................................................. */
 
static void cancel_find (GtkWidget *widget, gpointer data)
{
    gtk_widget_destroy(GTK_WIDGET(data));
    find_window = NULL;
}

/* .................................................................. */

static void find_string_dialog (void (*YesFunc)(), void (*NoFunc)(),
				gpointer data)
{
    GtkWidget *label;
    GtkWidget *button;
    GtkWidget *hbox;
    windata_t *mydat = (windata_t *) data;

    if (find_window) {
	gtk_object_set_data(GTK_OBJECT(find_window), "windat", mydat); 
	return;
    }

    find_window = gtk_dialog_new();
    gtk_object_set_data(GTK_OBJECT(find_window), "windat", mydat);

    gtk_signal_connect (GTK_OBJECT (find_window), "destroy",
	                GTK_SIGNAL_FUNC (close_find_dialog),
	                find_window);
    gtk_window_set_title (GTK_WINDOW (find_window), "gretl: find");
    gtk_container_border_width (GTK_CONTAINER (find_window), 5);

    hbox = gtk_hbox_new(TRUE, TRUE);
    label = gtk_label_new(" Find what:");
    gtk_widget_show (label);
    find_entry = gtk_entry_new();

    if (needle) {
	gtk_entry_set_text(GTK_ENTRY (find_entry), needle);
	gtk_entry_select_region (GTK_ENTRY (find_entry), 0, 
				 strlen (needle));
    }
    gtk_signal_connect(GTK_OBJECT (find_entry), 
			"activate", 
			GTK_SIGNAL_FUNC (YesFunc),
	                find_window);
    gtk_widget_show (find_entry);

    gtk_box_pack_start (GTK_BOX(hbox), label, TRUE, TRUE, 0);
    gtk_box_pack_start (GTK_BOX(hbox), find_entry, TRUE, TRUE, 0);
    gtk_widget_show (hbox);

    gtk_box_pack_start(GTK_BOX (GTK_DIALOG (find_window)->vbox), 
                        hbox, TRUE, TRUE, 0);

    gtk_box_set_spacing(GTK_BOX (GTK_DIALOG (find_window)->action_area), 15);
    gtk_box_set_homogeneous(GTK_BOX 
			     (GTK_DIALOG (find_window)->action_area), TRUE);
    gtk_window_set_position(GTK_WINDOW (find_window), GTK_WIN_POS_MOUSE);

    /* find button -- make this the default */
    button = gtk_button_new_with_label ("Find next");
    GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
    gtk_box_pack_start(GTK_BOX (GTK_DIALOG (find_window)->action_area), 
		       button, TRUE, TRUE, FALSE);
    gtk_signal_connect(GTK_OBJECT (button), "clicked",
		       GTK_SIGNAL_FUNC (YesFunc), find_window);
    gtk_widget_grab_default(button);
    gtk_widget_show(button);

    /* cancel button */
    button = gtk_button_new_with_label ("Cancel");
    GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
    gtk_box_pack_start(GTK_BOX (GTK_DIALOG (find_window)->action_area), 
		       button, TRUE, TRUE, FALSE);
    gtk_signal_connect(GTK_OBJECT (button), "clicked",
		       GTK_SIGNAL_FUNC (NoFunc), find_window);
    gtk_widget_show(button);

    gtk_widget_grab_focus(find_entry);
    gtk_widget_show (find_window);
}

/* .................................................................. */

void text_copy (gpointer data, guint all, GtkWidget *widget) 
{
    windata_t *mydata = (windata_t *) data;

    if (all) 
	gtk_editable_select_region (GTK_EDITABLE (mydata->w), 0, -1);
    gtk_editable_copy_clipboard (GTK_EDITABLE(mydata->w));
}

/* .................................................................. */

void make_menu_item (gchar *label, GtkWidget *menu,
		     GtkSignalFunc func, gpointer data)
{
    GtkWidget *item;

    item = gtk_menu_item_new_with_label(label);
    gtk_menu_append(GTK_MENU(menu), item);
    gtk_signal_connect_object(GTK_OBJECT(item), "activate",
			      GTK_SIGNAL_FUNC(func), data);
    gtk_widget_show(item);
}

/* .................................................................. */

void allocate_fileptrs (void)
{
    int i;
    
    for (i=0; i<MAXRECENT; i++) {
	datap[i] = datalist[i];
	sessionp[i] = sessionlist[i];
	scriptp[i] = scriptlist[i];
    }
}

/* .................................................................. */

void mkfilelist (int filetype, const char *fname)
{
    char *tmp[MAXRECENT-1];
    char **filep;
    int i, match = -1;

    if (filetype == 1) filep = datap;
    else if (filetype == 2) filep = sessionp;
    else if (filetype == 3) filep = scriptp;
    else return;

    /* see if this file is already on the list */
    for (i=0; i<MAXRECENT; i++) {
        if (strcmp(filep[i], fname) == 0) {
            match = i;
            break;
        }
    }
    if (match == 0) return; /* no change in list */
    else  { /* clear menu files list before rebuilding */
	GtkWidget *w;
	char tmpname[MAXSTR];
	gchar itempath[80];
	gchar *pathstart[] = {"/File/Open data", 
			      "/Session/Open",
			      "/File/Open command file"};

	for (i=0; i<MAXRECENT; i++) {
	    sprintf(itempath, "%s/%d. %s", pathstart[filetype - 1],
		    i+1, endbit(tmpname, filep[i], -1));
	    w = gtk_item_factory_get_widget(mdata->ifac, itempath);
	    if (w != NULL) 
		gtk_item_factory_delete_item(mdata->ifac, itempath);
	}
    }
    
    /* save pointers to current order */
    for (i=0; i<MAXRECENT-1; i++) tmp[i] = filep[i];

    /* copy fname into array, if not already present */
    if (match == -1) {
        for (i=1; i<MAXRECENT; i++) {
            if (filep[i][0] == '\0') {
                strcpy(filep[i], fname);
                match = i;
                break;
	    }
	    if (match == -1) {
		match = MAXRECENT - 1;
		strcpy(filep[match], fname);
	    }
	}
    } 
    /* set first pointer to new file */
    filep[0] = filep[match];

    /* rearrange other pointers */
    for (i=1; i<=match; i++) filep[i] = tmp[i-1];

    add_files_to_menu(filetype);
}

/* .................................................................. */

char *endbit (char *dest, char *src, int addscore)
{
    /* take last part of src filename */
    if (strrchr(src, SLASH))
	strcpy(dest, strrchr(src, SLASH) + 1);
    else
	strcpy(dest, src);

    if (addscore != 0) {
	/* then either double (1) or delete (-1) any underscores */
	char mod[MAXSTR];
	size_t i, j, n;

	n = strlen(dest);
	j = 0;
	for (i=0; i<=n; i++) {
	    if (dest[i] != '_')
		mod[j++] = dest[i];
	    else {
		if (addscore == 1) {
		    mod[j++] = '_';
		    mod[j++] = dest[i];
		} 
	    }
	}
	strcpy(dest, mod);
    }
    return dest;
}

/* .................................................................. */

static void printfilelist (int filetype, FILE *fp)
{
    int i;
    char **filep;

    if (filetype == 1) {
	fprintf(fp, "recent data files:\n");
	filep = datap;
    } else if (filetype == 2) {
	fprintf(fp, "recent session files:\n");
	filep = sessionp;
    } else if (filetype == 3) {
	fprintf(fp, "recent script files:\n");
	filep = scriptp;
    } else 
	return;

    for (i=0; i<MAXRECENT; i++) {
	if (filep[i][0]) 
	    fprintf(fp, "%s\n", filep[i]);
	else break;
    }
}

/* .................................................................. */

void add_files_to_menu (int filetype)
{
    int i;
    char **filep, tmp[MAXSTR];
    void (*callfunc)();
    GtkItemFactoryEntry fileitem;
    GtkWidget *w;
    gchar *msep[] = {"/File/Open data/sep",
		     "/Session/Open/sep",
		     "/File/Open command file/sep"};
    gchar *mpath[] = {"/File/_Open data",
		     "/Session/_Open",
		     "/File/Open command file"};

    fileitem.path = NULL;

    if (filetype == 1) {
	callfunc = set_data_from_filelist;
	filep = datap;
    } else if (filetype == 2) {
	callfunc = set_session_from_filelist;
	filep = sessionp;
    } else if (filetype == 3) {
	callfunc = set_script_from_filelist;
	filep = scriptp;
    }
    else
	return;

    /* See if there are any files to add */
    if (filep[0][0] == '\0') return;
    else {
	gchar *itemtype = "<Separator>";
	GtkWidget *w;

	/* is a separator already in place? */
	w = gtk_item_factory_get_widget(mdata->ifac, msep[filetype - 1]);
	if (w == NULL) {
	    fileitem.path = mymalloc(80);
	    strcpy(fileitem.path, mpath[filetype - 1]);
	    strcat(fileitem.path, "/sep");
	    fileitem.accelerator = NULL;
	    fileitem.callback = NULL;
	    fileitem.callback_action = 0;
	    fileitem.item_type = itemtype;
	    gtk_item_factory_create_item(mdata->ifac, &fileitem, NULL, 1);
	}
    }

    /* put the files under the menu separator */
    for (i=0; i<MAXRECENT; i++) {
	if (filep[i][0]) {
	    if (fileitem.path == NULL)
		fileitem.path = mymalloc(80);
	    fileitem.accelerator = NULL;
	    fileitem.callback_action = i; 
	    fileitem.item_type = NULL;
	    sprintf(fileitem.path, "%s/%d. %s", mpath[filetype - 1],
		    i+1, endbit(tmp, filep[i], 1));
	    fileitem.callback = callfunc; 
	    gtk_item_factory_create_item(mdata->ifac, &fileitem, NULL, 1);
	    w = gtk_item_factory_get_widget_by_action(mdata->ifac, i);
	    if (w != NULL)
		gtk_tooltips_set_tip(gretl_tips, w, filep[i], NULL);
	} else break;
    }
    free(fileitem.path);
}

/* .................................................................. */

#ifdef G_OS_WIN32
# include <windows.h>
#else
# include <dlfcn.h>
#endif

int open_plugin (const char *plugin, void **handle)
{
    char pluginpath[MAXLEN];

#ifdef G_OS_WIN32
    sprintf(pluginpath, "%s.dll", plugin);
    *handle = LoadLibrary(pluginpath);
    if (*handle == NULL) {
	errbox("Couldn't load plugin");
	return 1;
    }
#else
    sprintf(pluginpath, "%splugins/%s.so", paths.gretldir, plugin);
    *handle = dlopen(pluginpath, RTLD_LAZY);
    if (*handle == NULL) {
	errbox(dlerror());
	return 1;
    }
#endif 
    return 0;
}

/* .................................................................. */

void *get_plugin_function (const char *funcname, void *handle)
{
    void *funp;
    char *error;

#ifdef G_OS_WIN32
    funp = GetProcAddress(handle, funcname);
    if (funp == NULL)  {
	errbox("Couldn't access plugin function");
	return NULL;
    }
#else
    funp = dlsym(handle, funcname);
    if ((error = dlerror()) != NULL)  {
	errbox(error);
	return NULL;
    }
#endif   
    return funp;
}

/* .................................................................. */

void close_plugin (void *handle)
{
#ifdef G_OS_WIN32
    FreeLibrary(handle);
#else
    dlclose(handle);
#endif
}

